import jax.numpy as np
from jax import config
config.update("jax_enable_x64", True)
from jax.lax import *
from jax import custom_jvp,jvp
from functools import partial
from noloadj.numerical_methods.ODE.ode45 import next_der_step_simulation,\
    next_step_simulation,compute_new_point,odeint45,odeint45_etendu

def odeint45_extract(f,x0,*P,M=None,T=0.,h0=1e-5,tol=1.48e-8):
    '''
    Solves an ODE system described by f with Runge-Kutta 5 (7M) algorithm, without
    storing values across time and with features extraction.

    :param f: a class inherited from ODESystem abstract class that describes the ODE system
    :param x0: Array: initial state vector
    :param P: tuple : simulation parameters to optimize
    :param M: int: number of points to compute the FFT
    :param T: float: operating period of the system (optional)
    :param h0: float: initial step size
    :param tol: float: tolerance for optimal step size computation
    :return: Several outputs :
    - float : final time 'tf'
    - Array : final state vector 'xf'
    - Array : final output vector 'yf'
    - dict : extracted features dictionary 'cstr'
    - int/Array : final configuration 'state' if the system is periodic.
    '''
    return _odeint45_extract(f,h0,tol,M,x0,T,*P)

def compute_init_time_cstr(f,cstr,x0,y0,h0,dcstr=None,dx0=None,dy0=None):
    test_dcstr=not isinstance(dcstr,type(None)) # calcul initial des critères temporels
    for i in f.constraints.keys():
        if isinstance(f.constraints[i], tuple):
            test_exp,feature = f.constraints[i]
        else:
            feature = f.constraints[i]
            test_exp = lambda t: True
        if feature.type=='time':
            if feature.name in f.xnames:
                ind=f.xnames.index(feature.name)
                z0=x0
                if test_dcstr:
                    dz0=dx0
            else:
                ind=f.ynames.index(feature.name)
                z0=y0
                if test_dcstr:
                    dz0=dy0
            cstr[i]=np.where(test_exp(0.),feature.init(z0[ind],0.,h0),cstr[i])
            if test_dcstr:
                dcstr[i]=np.where(test_exp(0.),feature.dinit(z0[ind],dz0[ind],
                                    0.,h0),dcstr[i])
    if test_dcstr:
        return cstr,dcstr
    else:
        return cstr

def compute_expres_time_cstr(f,cstr,t_prev,x_prev,y_prev,h_prev,t_now,x_now,
        y_now,T,dcstr=None,dx_prev=None,dy_prev=None,dx_now=None,dy_now=None,
        dT=None):
    test_dcstr=not isinstance(dcstr,type(None)) # calcul pas à pas des critères temporels
    for i in f.constraints.keys():
        if isinstance(f.constraints[i], tuple):
            test_exp,feature= f.constraints[i]
        else:
            feature=f.constraints[i]
            test_exp = lambda t: True
        if feature.type=='time':
            if feature.name in f.xnames:
                ind=f.xnames.index(feature.name)
                z_prev,z_now=x_prev,x_now
                if test_dcstr:
                    dz_prev,dz_now=dx_prev,dx_now
            else:
                ind = f.ynames.index(feature.name)
                z_prev,z_now=y_prev,y_now
                if test_dcstr:
                    dz_prev,dz_now=dy_prev,dy_now
            cstr[i] =np.where(test_exp(t_now),feature.expression(t_prev,
                        z_prev[ind],t_now,z_now[ind], cstr[i],h_prev,T),cstr[i])
            if test_dcstr:
                dcstr[i]= np.where(test_exp(t_now),feature.der_expression(
                        t_prev,z_prev[ind],dz_prev[ind],t_now,z_now[ind],dz_now
                        [ind],cstr[i],dcstr[i],h_prev,T,dT),dcstr[i])
    if test_dcstr:
        return cstr, dcstr
    else:
        return cstr

def compute_fin_time_cstr(f,cstr,tf,xf,yf,T,dcstr=None,dT=None):
    test_dcstr=not isinstance(dcstr,type(None)) # itération finale des critères temporels
    for i in f.constraints.keys():
        if isinstance(f.constraints[i],tuple):
            _,feature=f.constraints[i]
        else:
            feature=f.constraints[i]
        if feature.type == 'time':
            if feature.name in f.xnames:
                ind = f.xnames.index(feature.name)
                zf=xf
            else:
                ind = f.ynames.index(feature.name)
                zf=yf
            cstr[i]=feature.fin(tf,cstr[i],T)
            if test_dcstr:
                dcstr[i]=feature.der_fin(tf,cstr[i],T,dcstr[i],dT,zf[ind])
    if test_dcstr:
        return cstr, dcstr
    else:
        return cstr

def compute_freq_cstr(f,cstr,modX,modY,phX,phY,vect_freq,T,dcstr=None,
                      dmodX=None,dmodY=None,dphX=None,dphY=None):
    test_dcstr=not isinstance(dcstr,type(None)) # calcul des critères fréquentiels
    for i in f.constraints.keys():
        if isinstance(f.constraints[i], tuple):
            _, feature = f.constraints[i]
        else:
            feature = f.constraints[i]
        if feature.type == 'freq':
            if feature.name in f.xnames:
                ind = f.xnames.index(feature.name)
                modZ,phZ=modX,phX
                if test_dcstr:
                    dmodZ, dphZ = dmodX, dphX
            else:
                ind = f.ynames.index(feature.name)
                modZ,phZ=modY,phY
                if test_dcstr:
                    dmodZ, dphZ = dmodY, dphY
            if test_dcstr:
                cstr[i],dcstr[i]=feature.der_expression(modZ[ind],
                        phZ[ind],dmodZ[ind], dphZ[ind],vect_freq,1/T)
            else:
                cstr[i]=feature.expression(modZ[ind],phZ[ind],vect_freq,1/T)
    if test_dcstr:
        return cstr, dcstr
    else:
        return cstr


@partial(custom_jvp,nondiff_argnums=(0,1,2,3))
def _odeint45_extract(f,h0,tol,M,x0,T,*P):

    def cond_fn(state): # condition d'arrêt de simulation
        x_prev2,x_prev,t_prev,_,h_prev,cstr,_,_=state
        return (h_prev > 0) & f.stop.cond_stop(t_prev,t_prev+h_prev,cstr,
                                               x_prev2,x_prev,T)

    def body_fn(state):
        _,x_prev,t_prev,y_prev,h_prev,cstr,i_prev,c_prev=state

        x_now,t_now,y_now,h_now,i_now,c_now=next_step_simulation(x_prev,t_prev,
                                    y_prev,i_prev,c_prev,h_prev,f,tol,T,*P) # prochaine itération de simulation

        if f.stop.type=='threshold': # si critère d'arrêt sous forme de franchissement de seuil,
            # interpolation linéaire sur le dernier point lorsque la simulation se termine
            output,seuil=x_now[f.stop.ind],f.stop.threshold
            outputprev=x_prev[f.stop.ind]
            condition=np.bitwise_not(f.stop.cond_stop(t_prev,t_now,cstr,x_prev,
                                                      x_now,T))
            x_now,h_prev,t_now,y_now=cond(condition,lambda state:
                compute_new_point(x_now,x_prev,t_prev,t_now,y_prev,y_now,
                output-seuil,outputprev-seuil),lambda state:state,(x_now,h_prev,
                                                                   t_now,y_now))
            x_now,t_now,y_now,h_now,i_now,c_now=cond(condition,lambda state:
                next_step_simulation(x_prev,t_prev,y_prev,i_prev,c_prev,h_prev,
                f,tol,T,*P),lambda state:state,(x_now,t_now,y_now,h_now,i_now,
                                                c_now))

        elif f.stop.type=='final_time': # atteindre point final
            h_tf=f.stop.tf-t_now
            h_now=np.minimum(h_now,h_tf)

        cstr=compute_expres_time_cstr(f,cstr,t_prev,x_prev,y_prev,h_prev,t_now,
                                      x_now,y_now,T) # calcul des critères temporels pas à pas

        return x_prev,x_now,t_now,y_now,h_now,cstr,i_now,c_now

    if f.stop.type=='steady_state':
        f.constraints=f.stop.init_dico(f.constraints,T)
    elif f.stop.type=='threshold':
        f.stop.init_ind(f.xnames)
    cstr=dict(zip(list(f.constraints.keys()),[0.]*len(f.constraints)))# INITIALISATION
    if not isinstance(f.state,type(None)): # si le systeme comporte plusieurs configurations
        i0=f.state
    else: # si le système ne comporte pas plusieurs configurations
        i0=0
    if not isinstance(f.commande,type(None)):# si le systeme comporte une commande
        c0=f.commande.next_command(0.)
    else:# si le systeme ne comporte pas une commande
        c0=0
    tempP=None
    if 'initialize' in f.__class__.__dict__:
        tempP=P
        P=f.initialize(*P)
    y0=f.output(x0,0.,*P)
    if isinstance(y0,tuple): # si des grandeurs d'état deviennent de sortie
        x0,y0=y0
    if 'initialize' in f.__class__.__dict__:
        P=tempP

    cstr=compute_init_time_cstr(f,cstr,x0,y0,h0)# initialisation des critères temporels

    _,xf,tf,yf,hf,cstr,ifinal,_=while_loop(cond_fn,body_fn,
                                         (x0,x0,0.,y0,h0,cstr,i0,c0))
    if not isinstance(f.state,type(None)): # si le systeme comporte plusieurs configurations
        f.state=ifinal

    cstr=compute_fin_time_cstr(f,cstr,tf,xf,yf,T) # iteration finale des critères temporels
    
    if f.stop.type=='steady_state':
        f.constraints=f.stop.delete_dico(f.constraints)
        cstr=f.stop.delete_dico(cstr)

    if M!=None: # s'il y a une simulation fréquentielle d'une période à faire après détection du régime permanent
        _,modX,phX,_,modY,phY,_=odeint45(f,xf,np.linspace(tf,tf+T,M),*P,
                                        M=M,T=T,h0=h0)
        vect_freq=np.where(M//2==0,np.linspace(0.,(M/2-1)/T,M//2),
                              np.linspace(0.,(M-1)/(2*T),M//2))

        cstr=compute_freq_cstr(f,cstr,modX,modY,phX,phY,vect_freq,T)

    if not isinstance(f.state,type(None)): # si le systeme comporte plusieurs configurations
        return (tf,xf,yf,cstr,ifinal)
    else: # si le système ne comporte pas plusieurs configurations
        return (tf,xf,yf,cstr)


@_odeint45_extract.defjvp
def _odeint45_extract_jvp(f,h0,tol,M, primals, tangents):
    '''
    Solves an ODE system described by df/dP with Runge-Kutta 5 (7M) algorithm, and
    computes derivatives of extracted features w.r.t optimization inputs.

    :param f: a class that describes the ODE system
    :param h0: initial step size
    :param tol: tolerance for optimal step size computation
    :param M: number of points to compute the FFT
    :param primals: tuple including initial state vector,operating period and
        optimization inputs.
    :param tangents: tuple including differentials of initial state vector,
        operating period and optimization inputs.
    :return: the final time tf, final state vector xf, final output vector yf,extracted features dictionary cstr, their derivatives w.r.t P, and the final
    configuration state if the system is periodic.
    '''
    x0,T, *P = primals
    dx0,dT, *dP = tangents
    nPdP = len(P)

    res=odeint45_extract_etendu(f,nPdP,h0, tol,M, x0,dx0,T,dT, *P, *dP)

    if not isinstance(f.state,type(None)):# si le systeme comporte plusieurs configurations
        xf,yf,cstr,tf,dtf,dxf,dyf,dcstr,states=res
        states=np.float64(states)
        return (tf,xf,yf,cstr,states),(dtf,dxf,dyf,dcstr,states)
    else:# si le système ne comporte pas plusieurs configurations
        xf,yf,cstr,tf,dtf,dxf,dyf,dcstr=res
        return (tf,xf,yf,cstr),(dtf,dxf,dyf,dcstr)


def odeint45_extract_etendu(f,nPdP,h0,tol,M,x0,dx0,T,dT,*P_and_dP):
    P,dP = P_and_dP[:nPdP],P_and_dP[nPdP:]

    def cond_fn(state): # condition d'arrêt de simulation
        x_prev2,x_prev,_,_,_,t_prev, h_prev,cstr,_,_,_,_,_,_ = state
        return (h_prev > 0) & f.stop.cond_stop(t_prev,t_prev+h_prev,cstr,
                                               x_prev2,x_prev,T)

    def body_fn(state):
        _,x_prev,dx_prev,y_prev,dy_prev,t_prev, h_prev,cstr,\
                dcstr,i_prev,c_prev,Mat,dMat,chgt_state= state

        x_now,t_now,y_now,h_now,i_now,c_now=next_step_simulation(x_prev,t_prev,
                                        y_prev,i_prev,c_prev,h_prev,f,tol,T,*P) # prochaine itération de simulation
        dx_now,dy_now,Mat,dMat=next_der_step_simulation(x_prev,t_prev,dx_prev,
                    x_now,t_now,h_prev,f, nPdP,chgt_state,Mat,dMat,*P_and_dP) # prochaine itération de calcul des gradients
        chgt_state = np.bitwise_not(np.array_equal(i_now, i_prev))

        if f.stop.type=='threshold': # si critère d'arrêt sous forme de franchissement de seuil,
            # interpolation linéaire sur le dernier point lorsque la simulation se termine
            output,seuil=x_now[f.stop.ind],f.stop.threshold
            outputprev=x_prev[f.stop.ind]
            condition=np.bitwise_not(f.stop.cond_stop(t_prev,t_now,cstr,x_prev,
                                                      x_now,T))
            x_now,h_prev,t_now,y_now=cond(condition,lambda state:
                compute_new_point(x_now,x_prev,t_prev,t_now,y_prev,y_now,
                output-seuil,outputprev-seuil),lambda state:state,(x_now,h_prev,
                                                                t_now,y_now))
            x_now,t_now,y_now,h_now,i_now,c_now=cond(condition,lambda state:
                next_step_simulation(x_prev,t_prev,y_prev,i_prev,c_prev,h_prev,
                f,tol,T,*P),lambda state:state,(x_now,t_now,y_now,h_now,i_now,
                                                c_now))
            dx_now,dy_now,_,_=cond(condition,lambda state:next_der_step_simulation
                (x_prev,t_prev,dx_prev,x_now,t_now,h_prev,f, nPdP,i_now,Mat,
                 dMat,*P_and_dP),lambda state:state,(dx_now,dy_now,Mat,dMat))

        elif f.stop.type=='final_time': # atteindre point final
            h_tf=f.stop.tf-t_now
            h_now=np.minimum(h_now,h_tf)

        cstr,dcstr=compute_expres_time_cstr(f,cstr,t_prev,x_prev,y_prev,h_prev,
            t_now,x_now,y_now,T,dcstr,dx_prev,dy_prev,dx_now,dy_now,dT) # calcul des critères temporels et de leurs gradients pas à pas

        return x_prev,x_now,dx_now,y_now,dy_now,t_now,h_now,cstr,dcstr,i_now,\
               c_now,Mat,dMat,chgt_state

    if f.stop.type=='steady_state':
        f.constraints=f.stop.init_dico(f.constraints,T)
    elif f.stop.type=='threshold':
        f.stop.init_ind(f.xnames)
    cstr=dict(zip(list(f.constraints.keys()),[0.]*len(f.constraints)))#INITIALISATION
    dcstr=dict(zip(list(f.constraints.keys()),[0.]*len(f.constraints)))

    for element in f.__dict__.keys(): # pour eviter erreurs de code
        if hasattr(f.__dict__[element],'primal'):
            f.__dict__[element]=f.__dict__[element].primal
    if not isinstance(f.state,type(None)): # si le systeme comporte plusieurs configurations
        i0=f.state
    else: # si le système ne comporte pas plusieurs configurations
        i0=0
    if not isinstance(f.commande,type(None)): # si le système ne comporte pas une commande
        c0=f.commande.next_command(0.)
    else: # si le système ne comporte pas de commande
        c0=0
    tempP,tempdP,Mat,dMat=None,None,0,0
    if 'initialize' in f.__class__.__dict__:
        tempP,tempdP=P,dP
        P,dP=jvp(f.initialize,(*P,),(*dP,))
        Mat,dMat=P,dP
    y0,dy0=jvp(f.output,(x0,0.,*P),(dx0,0.,*dP))
    if isinstance(y0,tuple): # si des grandeurs d'état deviennent de sortie
        x0,y0=y0
        dx0,dy0=dy0
    if 'initialize' in f.__class__.__dict__:
        P,dP=tempP,tempdP

    cstr,dcstr=compute_init_time_cstr(f,cstr,x0,y0,h0,dcstr,dx0,dy0) # initialisation des critères temporels
        # et de leurs gradients

    chgt_state=False
    xfm1,xf,dxf,yf,dyf,tf,hf,cstr,dcstr,ifinal,_,_,_,_=while_loop(cond_fn,
        body_fn,(x0,x0,dx0,y0,dy0,0.,h0,cstr,dcstr,i0,c0,Mat,dMat,chgt_state))
    if not isinstance(f.state,type(None)): # si le systeme comporte plusieurs configurations
        f.state=ifinal
    if 'initialize' in f.__class__.__dict__:
        P=f.initialize(*P)

    cstr,dcstr=compute_fin_time_cstr(f,cstr,tf,xf,yf,T,dcstr,dT)# iteration finale des critères temporels

    if f.stop.type=='threshold': # derivees partielles de tf, cas franchissement de seuil
        dout=dxf[f.stop.ind]
        xseuil=f.timederivatives(xf,tf,*P)[f.stop.ind]
        dtf=-(1/xseuil)*dout
    elif f.stop.type=='steady_state':
        ind_rp=0
        xseuil=f.timederivatives(xf,tf,*P)[ind_rp]
        dtf=-(1/xseuil)*dxf[ind_rp]
    else: # derivees partielles de tf, cas date de fin imposée
        dtf=0.
        
    if f.stop.type=='steady_state':
        f.constraints=f.stop.delete_dico(f.constraints)
        cstr=f.stop.delete_dico(cstr)
        dcstr=f.stop.delete_dico(dcstr)

    if M!=None: # s'il y a une simulation fréquentielle d'une période à faire après détection du régime permanent
        _,_,modX,phX,dmodX,dphX,_,_,modY,phY,dmodY,dphY,_=odeint45_etendu(f,
            nPdP,h0,tol,M,xf,dxf,np.linspace(tf,tf+T,M),T,*P_and_dP)

        vect_freq=np.where(M//2==0,np.linspace(0.,(M/2-1)/T,M//2),
                              np.linspace(0.,(M-1)/(2*T),M//2))
        cstr,dcstr=compute_freq_cstr(f,cstr,modX,modY,phX,phY,vect_freq,T,dcstr,
                      dmodX,dmodY,dphX,dphY)

    if not isinstance(f.state,type(None)):  # si le systeme comporte plusieurs configurations
        return xf,yf,cstr,tf,dtf,dxf,dyf,dcstr,ifinal
    else: # si le système ne comporte pas plusieurs configurations
        return xf,yf,cstr,tf,dtf,dxf,dyf,dcstr


from noloadj.numerical_methods.ODE.ode_extracted_features import T_pair,\
    T_impair,Min,Max
from abc import abstractmethod

class StoppingCriteria:
    def __init__(self):
        self.type=''
    @abstractmethod
    def cond_stop(self,t_prev,t,cstr,x_prev,x,T): # condition d'arrêt
        pass

class threshold(StoppingCriteria): # critère d'arrêt sous franchissement de seuil d'une variable
    def __init__(self,var,threshold_=0.):
        StoppingCriteria.__init__(self)
        self.type='threshold'
        self.var=var
        self.threshold=threshold_
    def init_ind(self,names_var):
        self.ind=names_var.index(self.var)
    def cond_stop(self,t_prev,t,cstr,x_prev,x,T): # condition d'arrêt
        return np.sign(x[self.ind]-self.threshold)==np.sign(x_prev[self.ind]-
                                                            self.threshold)

class final_time(StoppingCriteria): # date de fin de simulation imposée
    def __init__(self,tf):
        StoppingCriteria.__init__(self)
        self.type='final_time'
        self.tf=tf
    def cond_stop(self,t_prev,t,cstr,x_prev,x,T): # condition d'arrêt
        return t_prev<self.tf

class steady_state(StoppingCriteria): # detection de régime permanent

    def __init__(self,nbT=0,names_var=[],a=1e-5):
        StoppingCriteria.__init__(self)
        self.type='steady_state'
        self.nbT=nbT
        self.names_var=names_var
        self.a=a
    def init_dico(self,constr,T):
        if self.nbT!=0:
            for i in range(len(self.names_var)):
                constr[self.names_var[i]+'_minp']=(T_pair(self.nbT * T),
                                                Min(self.names_var[i],self.nbT))
                constr[self.names_var[i]+'_minimp']=(T_impair(self.nbT * T),
                                                Min(self.names_var[i],self.nbT))
                constr[self.names_var[i]+'_maxp']=(T_pair(self.nbT * T),
                                                Max(self.names_var[i],self.nbT))
                constr[self.names_var[i]+'_maximp']=(T_impair(self.nbT * T),
                                                Max(self.names_var[i],self.nbT))
        return constr
    def cond_stop(self,t_prev,t,cstr,x_prev,x,T): # condition d'arrêt
        if self.nbT!=0:
            vectp,vectimp=np.zeros(2*len(self.names_var)),np.zeros(2*len(
                self.names_var))
            for i in range(len(self.names_var)):
                vectp=vectp.at[i].set(cstr[self.names_var[i]+'_minp'])
                vectp=vectp.at[2*i+1].set(cstr[self.names_var[i]+'_maxp'])
                vectimp=vectimp.at[i].set(cstr[self.names_var[i]+'_minimp'])
                vectimp=vectimp.at[2*i+1].set(cstr[self.names_var[i]+'_maximp'])
            return np.bitwise_not(np.bitwise_and(np.allclose(vectp,vectimp,
                                    atol=self.a),np.not_equal(t_prev//T,t//T)))
        else:
            return np.bitwise_not(np.bitwise_and(np.allclose(x_prev,x,
                                        atol=self.a),np.not_equal(t_prev,0.)))
    def delete_dico(self,constr):
        if self.nbT!=0:
            for i in range(len(self.names_var)):
                del constr[self.names_var[i]+'_minp']
                del constr[self.names_var[i]+'_minimp']
                del constr[self.names_var[i]+'_maxp']
                del constr[self.names_var[i]+'_maximp']
        return constr


import jax.numpy as np
from jax.lax import switch,cond
from jax import custom_jvp,jvp,Array
from functools import partial
from typing import Callable
from noloadj.numerical_methods.ODE.ode45_extract import final_time
import matplotlib.pyplot as plt

def vect_temporel(debut,fin,pas):
    '''
    Creates a time vector.

    :param debut: float: starting time
    :param fin: float: ending time
    :param pas: float : step size
    :return: JaxArray: time vector
    '''
    return np.linspace(debut,fin,int((fin-debut)/pas))

def indice_t(t,pas,debut=0.):
    return ((t-debut)/pas).astype(int)

def Switch(etat,funcs):
    return switch(etat,[func for func in funcs])

def Condition(conditions,functions,state):
    for i in range(len(conditions)):
        state=cond(conditions[i],functions[i],lambda state:state,state)
    return state


###################################################### integrale
@partial(custom_jvp,nondiff_argnums=(0,1,2,3))
def integrale(t0,tf,pas,f,*inputs):
    '''
    Computes the integral of a function on a domain.

    :param t0: float: initial bound of the integral
    :param tf: float: final bounds of the integral
    :param pas: float: step size of the time vector
    :param f: the function on which the integral has to be computed
    :param inputs: input variables across a time vector
    :return: the integral value
    '''
    longueur=(tf-t0)/int(tf/pas)
    res=f(*inputs)
    S=2*np.sum(res)-res[0]-res[-1]
    S*=longueur*0.5
    return S

@integrale.defjvp
def integrale_jvp(t0,tf,pas,f,primals,tangents):
    '''
    Computes the derivatives of the integral of a function on a domain w.r.t inputs.

    :param t0: float: initial bound of the integral
    :param tf: float: final bounds of the integral
    :param pas: float: step size of the time vector
    :param f: the function on which the integral has to be computed
    :param primals: tuple of input values
    :param tangents: tuple of input differentials
    :return: JaxArray: Derivatives of integral w.r.t inputs
    '''
    longueur=(tf-t0)/int(tf/pas)
    primal_dot,tangent_dot=jvp(f,primals,tangents)
    S=2*np.sum(primal_dot)-primal_dot[0]-primal_dot[-1]
    S*=longueur*0.5
    dS = 2*np.sum(tangent_dot)-tangent_dot[0]-tangent_dot[-1]
    dS *= longueur * 0.5
    return S,dS
############################################################################

def vect_freq(t0,tf,Te):
    '''
    Computes the frequency vector for FFT computation.

    :param t0: initial time
    :param tf: final time
    :param Te: sampling period
    :return: the frequency vector
    '''
    M=int((tf-t0)/Te)
    freq=np.where(M//2==0,np.linspace(0.,(M/2-1)/(M*Te),M//2),
                          np.linspace(0.,(M-1)/(2*M*Te),M//2))
    return freq

class Creneau:
    '''
    Class to implement a MLI command of the system (optional)
    
    Attributes:
    - a : float for Duty cycle of the system
    - T : float for Operating period of the system
    '''
    def __init__(self,a,T):
        self.aT=a*T
        self.T=T
        if hasattr(self.aT,'primal'):
            self.aT=self.aT.primal
        if hasattr(self.T,'primal'):
            self.T=self.T.primal
    
    def next_command(self,t):
        '''
        Returns command value for next iteration.
        :param t: float :time at present iteration
        :return: command value for next iteration : c
        '''
        moduloT=(t//self.T)*self.T
        c=np.where(t-moduloT<self.aT,1,0)
        return c
        
    def next_pdi(self,t):
        '''
        Returns next important date. tpdi.
        :param t: float :time at present iteration
        :return: next important date tpdi.
        '''
        moduloT=(t//self.T)*self.T
        tpdi=np.where(t-moduloT<self.aT,self.aT+moduloT,self.T+moduloT)
        return tpdi+1e-12
        

from abc import abstractmethod

class ODESystem:
    '''
    Interface class to call when you create a dynamic system class.

    Attributes:
        - xnames : list for state variables names
        - ynames : list for output variables names
        - stop : Callable for stopping criteria
        - constraints : dict for features to extract from dynamic simulation
        - state: int/Array: configuration of the system (optional)
        - commande : Callable to call a MLI command of the system (optional)
    '''
    def __init__(self):
        self.xnames:list=[]
        self.ynames:list=[]
        self.stop:Callable=final_time(1.0) # initialization with a stopping criteria
        self.constraints : dict={}
        self.state:[int,Array]=None
        self.commande:Callable=None # Class to implement a MLI command of the system (optional)

    @abstractmethod
    def timederivatives(self,X:Array,t:float,*P):
        '''
        Method to compute time derivatives of state vector.

        :param X: Array : state vector at present iteration
        :param t: float :time at present iteration
        :param P: tuple : simulation parameters to optimize
        :return: Array : time derivatives of state vector.
        '''
        pass

    @abstractmethod
    def output(self,X:Array,t:float,*P):
        '''
        Method to compute output vector.

        :param X: Array : state vector at present iteration
        :param t: float :time at present iteration
        :param P: tuple : simulation parameters to optimize
        :return: Array : output vector
        '''
        pass

    @abstractmethod
    def update(self,X:Array,Y:Array,t:float):
        '''
        Method to do configuration changes of the system (optional).

        :param X: Array : state vector at present iteration
        :param Y: Array : output vector
        :param t: float :time at present iteration
        :return: Several outputs :
        - int : configuration at next iteration 'i_now' (optional)
        - Array : modified state vector Xm
        - Array : output vector Ym
        '''
        pass

    @abstractmethod
    def initialize(self,*P):
        '''
        Method to compute state matrices that changed during simulation process (optional).

        :param P: tuple : simulation parameters to optimize
        :return: state matrices
        '''
        pass

    def get_indice(self,xy,desired_var):
        '''
        Returns the value(s) of an array corresponding to the variable name(s)
        given in desired_names.

        :param xy: Array: the state/output vector
        :param desired_var: list: the name of the variable(s) to get
        :return: Array: the value(s) of the desired names of variables in an array.
        '''
        if desired_var[0] in self.xnames:
            names_vars=self.xnames
        else:
            names_vars=self.ynames
        if len(desired_var)==1:
            return xy[names_vars.index(desired_var[0])]
        else:
            return (xy[names_vars.index(i)] for i in desired_var)

    def plotvar(self,xy,time,desired_var=None):
        '''
        Plots evolution of state/output variables given in desired_names.

        :param xy: Array: the state/output vector
        :param time: Array : time vector
        :param desired_var: list: the name of the variable(s) to get
        :return: /
        '''
        plt.figure()
        if len(xy)==len(self.xnames):
            if desired_var==None:
                desired_var=self.xnames
            title='State variables vs time'
        elif len(xy)==len(self.ynames):
            if desired_var==None:
                desired_var=self.ynames
            title='Outputs vs time'
        else:
            title=''
        for element in desired_var:
            var=self.get_indice(xy,[element])
            plt.plot(time,var)
        plt.xlabel('t(s)')
        plt.legend(desired_var)
        plt.title(title)
        plt.show()


# SPDX-FileCopyrightText: 2021 G2Elab / MAGE
#
# SPDX-License-Identifier: Apache-2.0

from typing import List
import pandas as pd

class Solution:
    """
    Class used to store optimization inputs and outputs for an iteration.

    Attributes :

    - iData = list : model inputs (optimization variables) values
    - oData = list : model outputs (objective functions + constraints) values
    - fData = list : free outputs values
    """
    # iData = []
    # oData = []
    # fData = []

    def __init__(self, inp, out, freeOutputs=[]):
        self.iData = inp
        self.oData = out
        self.fData = freeOutputs



class Iterations:
    """
    Class used to store optimization inputs and outputs for each iteration.

    Attributes :

    - solutions: class Solution
    - iNames = list : model inputs
    - oNames = list : model outputs
    - fNames = list : model free outputs names
    - objectives = dict : objective functions of the model
    - eq_cstr = dict : equality constraints of the model
    - ineq_cstr = dict : inequality constraints of the model
    - bounds = dict : range of values of inputs
    """
    # solutions: Solution = []
    # iNames = [] # noms des entrées du modèle
    # oNames = [] # noms des sorties du modèle
    # fNames = [] # noms des sorties 'libres' du modèle
    # objectives = {} # fonctions objectives du modèle
    # eq_cstr = {} # contraintes d'égalité
    # ineq_cstr = {} # contraintes d'inégalité
    # bounds = {} # domaine de recherche

    def __init__(self, spec,iNames, oNames, fNames=[], handler = None):
        self.iNames = iNames
        self.oNames = oNames
        self.fNames = fNames
        self.solutions = []
        if spec!=None:
            if len(spec.objectives)==1:
                self.objectives=dict(zip(spec.objectives,[spec.objectives_val]))
            else:
                self.objectives=dict(zip(spec.objectives,spec.objectives_val))
            self.eq_cstr=spec.dict_eq_cstr
            self.ineq_cstr=spec.dict_ineq_cstr
            self.bounds=spec.dict_bounds
        self.handler = handler #TODO : n'est plus utilisé. Modifier la création
        # automatique du Handler (dans constructeur Wrapper) lorsqu'il s'agit
        # d'un affichage dynamique : dynamicPlot.update

    def updateData(self, inp, out, freeOutputs=[]):
        """
        Adds the inputs and outputs of the model computed at each iteration to
        the Solution class.

        :param inp: list of model inputs
        :param out: list of model outputs
        :param freeOutputs: list of freeOutputs (optional)
        :return: /
        """
        self.solutions.append(Solution(inp.copy(), out.copy(),
                                       freeOutputs.copy()))
        if (self.handler):
            self.handler(self.solutions)

    def print(self):
        """
        Returns the optimization variables values for each iteration of the optimization problem.

        :return: a dataframe with the optimization variables values for each iteration of the optimization problem.
        """
        liste=[]
        res_inp=[sol.iData for sol in self.solutions]
        res_out=[sol.oData for sol in self.solutions]
        if self.fNames !=[]:
            res_fout=[sol.fData for sol in self.solutions]
        for i in range(len(self.solutions)):
            dict={}
            for j in range(len(self.iNames)):
                dict[self.iNames[j]]=res_inp[i][j]
            for k in range(len(self.oNames)):
                dict[self.oNames[k]]=res_out[i][k]
            if self.fNames!=[]:
                for l in range(len(self.fNames)):
                    dict[self.fNames[l]]=res_fout[i][l]
            liste.append(dict)
        return pd.DataFrame(liste)
    def printSpec(self):
        """
        Returns the specifications of the optimization problem with a dataframe.

        :return: a dataframe including the specifications of the optimization problem.
        """
        df=pd.DataFrame(columns=["Value","In_Out","Type"],
                        index=list(set(self.iNames+self.oNames+self.fNames)))
        df.index.name = 'Name'
        for i in list(self.bounds.keys()):
            df.loc[i,"Value"] = self.bounds[i]
            df.loc[i,"In_Out"] = 'Input'
            df.loc[i,"Type"] = 'bounds'
        for i in list(self.objectives.keys()):
            df.loc[i,"Value"] = self.objectives[i]
            df.loc[i,"In_Out"] = 'Output'
            df.loc[i,"Type"] = 'objective'
        if self.eq_cstr!={}:
            for i in list(self.eq_cstr.keys()):
                df.loc[i,"Value"] = self.eq_cstr[i]
                df.loc[i,"In_Out"] = 'Output'
                df.loc[i,"Type"] = 'eq_cstr'
        if self.ineq_cstr!={}:
            for i in list(self.ineq_cstr.keys()):
                df.loc[i,"Value"] = self.ineq_cstr[i]
                df.loc[i,"In_Out"] = 'Output'
                df.loc[i,"Type"] = 'ineq_cstr'
        if self.fNames!=[]:
            for i in self.fNames:
                df.loc[i,"In_Out"] = 'Output'
                df.loc[i,"Type"] = 'free'
        return df

    def plotXY(self):
        import noloadj.gui.plotIterations as pi
        pi.plotXY(self)
    def plotIO(self):
        import noloadj.gui.plotIterations as pi
        pi.plotIO(self)


def printHandler(sols:List[Solution]):
    """
    Displays the inputs and outputs of the model computed during the last
    iteration.

    :param sols: class Solution
    :return: /
    """
    sols
    print(sols[-1].iData)
    print(sols[-1].oData)
    if sols.fData != []:
        print(sols[-1].fData)



